#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import glob
import logging
import sys
from pathlib import Path
import pbr.version
import webcolors

try:
    from gnuradio import analog  # pytype: disable=import-error
    from gnuradio import filter as grfilter  # pytype: disable=import-error
    from gnuradio import blocks  # pytype: disable=import-error
    from gnuradio import fft  # pytype: disable=import-error
    from gnuradio import gr  # pytype: disable=import-error
    from gnuradio import zeromq  # pytype: disable=import-error
    from gnuradio.fft import window  # pytype: disable=import-error
except ModuleNotFoundError as err:  # pragma: no cover
    print(
        "Run from outside a supported environment, please run via Docker (https://github.com/IQTLabs/gamutRF#readme): %s"
        % err
    )
    sys.exit(1)

from gamutrf.grsource import get_source
from gamutrf.grinferenceoutput import inferenceoutput
from gamutrf.grpduzmq import pduzmq
from gamutrf.utils import endianstr


class grscan(gr.top_block):
    def __init__(
        self,
        bucket_range=1.0,
        colormap=16,
        compass=False,
        db_clamp_ceil=50,
        db_clamp_floor=-200,
        dc_block_len=0,
        dc_block_long=False,
        dc_ettus_auto_offset=True,
        description="",
        external_gps_server="",
        external_gps_server_port=8888,
        fft_batch_size=256,
        fft_processor_affinity=0,
        iq_zmq_addr="0.0.0.0",
        iq_zmq_port=10002,
        freq_end=1e9,
        freq_start=100e6,
        gps_server="",
        igain=0,
        inference_addr="0.0.0.0",  # nosec
        inference_batch=1,
        inference_min_confidence=0.5,
        inference_min_db=-200,
        inference_model_name="",
        inference_model_server="",
        inference_output_dir="",
        inference_port=10001,
        inference_text_color="",
        iq_inference_background=True,
        iq_inference_model_name="",
        iq_inference_model_server="",
        iq_inference_squelch_db=None,
        iq_inference_squelch_alpha=1e-4,
        iq_power_inference=False,
        iqtlabs=None,
        fft_zmq_addr="0.0.0.0",  # nosec
        fft_zmq_port=10000,
        low_power_hold_down=False,
        mqtt_server="",
        n_image=0,
        n_inference=0,
        nfft=1024,
        peak_fft_range=0,
        pretune=False,
        rotate_secs=0,
        samp_rate=4.096e6,
        sample_dir="",
        scaling="spectrum",
        sdr="ettus",
        sdrargs=None,
        sigmf=True,
        skip_tune_step=0,
        slew_rx_time=True,
        sweep_sec=30,
        tag_now=False,
        tune_dwell_ms=0,
        tune_jitter_hz=0,
        tune_step_fft=0,
        tuneoverlap=0.5,
        tuning_ranges="",
        use_external_gps=False,
        use_external_heading=False,
        vkfft=False,
        wavelearner=None,
        write_fft_points=False,
        write_samples=0,
    ):
        gr.top_block.__init__(self, "scan", catch_exceptions=True)

        if description:
            description = description.strip('"')
        tune_step_hz = int(samp_rate * tuneoverlap)
        stare = False
        initial_freq = freq_start

        if freq_end == 0:
            stare = True
            freq_end = freq_start + (tune_step_hz - 1)
            initial_freq += int((freq_end - freq_start) / 2)
            logging.info(
                f"using stare mode, scan from {freq_start/1e6}MHz to {freq_end/1e6}MHz"
            )

        ##################################################
        # Parameters
        ##################################################
        self.freq_end = freq_end
        self.freq_start = freq_start
        self.sweep_sec = sweep_sec
        self.nfft = nfft
        self.wavelearner = wavelearner
        self.iqtlabs = iqtlabs
        self.samp_rate = samp_rate
        self.retune_pre_fft = None
        self.tag_now = tag_now

        ##################################################
        # Blocks
        ##################################################
        if iqtlabs is not None:
            griqtlabs_path = os.path.realpath(
                glob.glob("/usr/local/**/libgnuradio-iqtlabs.so", recursive=True)[0]
            )
            pbr_version = pbr.version.VersionInfo("gamutrf").version_string()
            logging.info(f"gamutrf {pbr_version} with gr-iqtlabs {griqtlabs_path}")

        logging.info(f"will scan from {freq_start} to {freq_end}")
        freq_range = freq_end - freq_start
        fft_rate = int(samp_rate / nfft)

        if not tune_step_fft:
            if tune_dwell_ms:
                tune_step_fft = int(fft_rate * (tune_dwell_ms / 1e3))
            else:
                target_retune_hz = freq_range / self.sweep_sec / tune_step_hz
                tune_step_fft = int(fft_rate / target_retune_hz)
                logging.info(
                    f"retuning across {freq_range/1e6}MHz in {self.sweep_sec}s, requires retuning at {target_retune_hz}Hz in {tune_step_hz/1e6}MHz steps ({tune_step_fft} FFTs)"
                )
        if not tune_step_fft:
            logging.info("tune_step_fft cannot be 0 - defaulting to nfft")
            tune_step_fft = nfft
        tune_dwell_ms = tune_step_fft / fft_rate * 1e3
        logging.info(
            f"requested retuning across {freq_range/1e6}MHz every {tune_step_fft} FFTs, dwell time {tune_dwell_ms}ms"
        )
        if stare and tune_dwell_ms > 1e3:
            logging.warn(">1s dwell time in stare mode, updates will be slow!")
        peak_fft_range = min(peak_fft_range, tune_step_fft)

        self.sources, cmd_port, self.workaround_start_hook = get_source(
            sdr,
            samp_rate,
            igain,
            nfft,
            tune_step_fft,
            agc=False,
            center_freq=initial_freq,
            sdrargs=sdrargs,
            dc_ettus_auto_offset=dc_ettus_auto_offset,
        )

        fft_batch_size, self.fft_blocks = self.get_fft_blocks(
            samp_rate,
            tune_jitter_hz,
            vkfft,
            fft_batch_size,
            nfft,
            freq_start,
            freq_end,
            tune_step_hz,
            tune_step_fft,
            skip_tune_step,
            tuning_ranges,
            pretune,
            fft_processor_affinity,
            low_power_hold_down,
            slew_rx_time,
        )
        self.fft_blocks = (
            self.get_dc_blocks(dc_block_len, dc_block_long)
            + self.fft_blocks
            + self.get_db_blocks(nfft, samp_rate, scaling)
        )
        self.last_db_block = self.fft_blocks[-1]
        fft_dir = ""
        self.samples_blocks = []
        self.write_samples_block = None
        if write_samples:
            if write_fft_points:
                fft_dir = sample_dir
            Path(sample_dir).mkdir(parents=True, exist_ok=True)
            samples_vlen = fft_batch_size * nfft
            self.samples_blocks.extend(
                [
                    # blocks.vector_to_stream(
                    #    gr.sizeof_gr_complex, fft_batch_size * nfft
                    # ),
                    # blocks.complex_to_interleaved_short(False, 32767),
                    # blocks.stream_to_vector(gr.sizeof_short, nfft * 2),
                    self.iqtlabs.write_freq_samples(
                        "rx_freq",
                        gr.sizeof_gr_complex,
                        "_".join(("cf32", endianstr())),
                        samples_vlen,
                        sample_dir,
                        "samples",
                        int(write_samples / samples_vlen),
                        skip_tune_step,
                        int(samp_rate),
                        rotate_secs,
                        igain,
                        sigmf,
                        zstd=True,
                        rotate=False,
                        description=description,
                    ),
                ]
            )
            self.write_samples_block = self.samples_blocks[-1]

        retune_fft = self.iqtlabs.retune_fft(
            tag="rx_freq",
            nfft=nfft,
            samp_rate=int(samp_rate),
            tune_jitter_hz=int(tune_jitter_hz),
            freq_start=int(freq_start),
            freq_end=int(freq_end),
            tune_step_hz=tune_step_hz,
            tune_step_fft=tune_step_fft,
            skip_tune_step_fft=skip_tune_step,
            fft_min=db_clamp_floor,
            fft_max=db_clamp_ceil,
            sdir=fft_dir,
            write_step_fft=write_samples,
            bucket_range=bucket_range,
            tuning_ranges=tuning_ranges,
            description=description,
            rotate_secs=rotate_secs,
            pre_fft=pretune,
            tag_now=self.tag_now,
            low_power_hold_down=(not pretune and low_power_hold_down),
            slew_rx_time=slew_rx_time,
            peak_fft_range=peak_fft_range,
        )
        self.fft_blocks.append(retune_fft)
        fft_zmq_block_addr = f"tcp://{fft_zmq_addr}:{fft_zmq_port}"
        self.pduzmq_block = pduzmq(fft_zmq_block_addr)
        logging.info("serving FFT on %s", fft_zmq_block_addr)

        if iq_zmq_port:
            iq_zmq_block_addr = f"tcp://{iq_zmq_addr}:{iq_zmq_port}"
            logging.info("serving I/Q samples and tags on %s", iq_zmq_block_addr)
            iq_zmq_block = zeromq.pub_sink(
                gr.sizeof_gr_complex,
                fft_batch_size * nfft,
                iq_zmq_block_addr,
                100,
                True,
                65536,
                "",
            )
            self.connect((self.retune_pre_fft, 0), (iq_zmq_block, 0))

        self.inference_blocks = []
        self.inference_output_block = None
        self.image_inference_block = None
        self.iq_inference_block = None

        if inference_output_dir:
            Path(inference_output_dir).mkdir(parents=True, exist_ok=True)

        if inference_text_color:
            wc = webcolors.name_to_rgb(inference_text_color, "css3")
            inference_text_color = ",".join(
                [str(c) for c in [wc.blue, wc.green, wc.red]]
            )

        if (inference_model_server and inference_model_name) or inference_output_dir:
            x = 640
            y = 640
            self.image_inference_block = self.iqtlabs.image_inference(
                tag="rx_freq",
                vlen=nfft,
                x=x,
                y=y,
                image_dir=inference_output_dir,
                convert_alpha=255,
                norm_alpha=0,
                norm_beta=1,
                norm_type=32,  # cv::NORM_MINMAX = 32
                colormap=colormap,  # cv::COLORMAP_VIRIDIS = 16, cv::COLORMAP_TURBO = 20,
                interpolation=1,  # cv::INTER_LINEAR = 1,
                flip=0,  # 0 means flipping around the x-axis
                min_peak_points=inference_min_db,
                model_server=inference_model_server,
                model_names=inference_model_name,
                confidence=inference_min_confidence,
                max_rows=tune_step_fft,
                rotate_secs=rotate_secs,
                n_image=n_image,
                n_inference=n_inference,
                samp_rate=int(samp_rate),
                text_color=inference_text_color,
            )
            self.inference_blocks.append(self.image_inference_block)

        if iq_inference_model_server and iq_inference_model_name:
            self.iq_inference_block = iqtlabs.iq_inference(
                tag="rx_freq",
                vlen=nfft,
                n_vlen=1,
                sample_buffer=tune_step_fft,
                min_peak_points=inference_min_db,
                model_server=iq_inference_model_server,
                model_names=iq_inference_model_name,
                confidence=inference_min_confidence,
                n_inference=n_inference,
                samp_rate=int(samp_rate),
                power_inference=iq_power_inference,
                background=iq_inference_background,
                batch=inference_batch,
            )
            self.inference_blocks.append(self.iq_inference_block)
            if self.write_samples_block:
                self.msg_connect(
                    (self.iq_inference_block, "inference"),
                    (self.write_samples_block, "inference"),
                )

        # TODO: provide new block that receives JSON-over-PMT and outputs to MQTT/zmq.
        retune_fft_output_block = None
        if self.inference_blocks:
            inference_zmq_addr = f"tcp://{inference_addr}:{inference_port}"
            self.inference_output_block = inferenceoutput(
                "inferencemqtt",
                inference_zmq_addr,
                mqtt_server,
                compass,
                gps_server,
                use_external_gps,
                use_external_heading,
                external_gps_server,
                external_gps_server_port,
                inference_output_dir,
            )
            if self.iq_inference_block:
                if iq_inference_squelch_db is not None:
                    squelch_blocks = [
                        blocks.vector_to_stream(
                            gr.sizeof_gr_complex,
                            fft_batch_size * nfft,
                        ),
                        analog.pwr_squelch_cc(
                            iq_inference_squelch_db,
                            iq_inference_squelch_alpha,
                            0,
                            False,
                        ),
                        blocks.stream_to_vector(
                            gr.sizeof_gr_complex,
                            fft_batch_size * nfft,
                        ),
                        self.iq_inference_block,
                    ]
                    self.connect_blocks(self.retune_pre_fft, squelch_blocks)
                else:
                    self.connect((self.retune_pre_fft, 0), (self.iq_inference_block, 0))
                self.connect((self.last_db_block, 0), (self.iq_inference_block, 1))
            if self.image_inference_block:
                if stare:
                    self.connect(
                        (self.last_db_block, 0), (self.image_inference_block, 0)
                    )
                else:
                    retune_fft_output_block = self.image_inference_block
            for block in self.inference_blocks:
                self.msg_connect(
                    (block, "inference"), (self.inference_output_block, "inference")
                )

        if retune_fft_output_block:
            self.connect((retune_fft, 0), (retune_fft_output_block, 0))

        if pretune:
            self.msg_connect((self.retune_pre_fft, "tune"), (self.sources[0], cmd_port))
            self.msg_connect((self.retune_pre_fft, "tune"), (retune_fft, "cmd"))
        else:
            self.msg_connect((retune_fft, "tune"), (self.sources[0], cmd_port))
        self.msg_connect((retune_fft, "json"), (self.pduzmq_block, "json"))
        self.connect_blocks(self.sources[0], self.sources[1:])

        self.connect_blocks(self.sources[-1], self.fft_blocks)
        self.connect_blocks(self.retune_pre_fft, self.samples_blocks)

    def connect_blocks(self, source, other_blocks, last_block_port=0):
        last_block = source
        for block in other_blocks:
            self.connect((last_block, last_block_port), (block, 0))
            last_block = block

    def get_db_blocks(self, nfft, samp_rate, scaling):
        if scaling == "density":
            scale = 1.0 / (samp_rate * sum(([x**2 for x in self.get_window(nfft)])))
        elif scaling == "spectrum":
            scale = 1.0 / (sum(self.get_window(nfft)) ** 2)
        else:
            raise ValueError("scaling must be 'spectrum' or 'density'")
        return [
            blocks.complex_to_mag_squared(nfft),
            blocks.multiply_const_ff(scale, nfft),
            blocks.nlog10_ff(10, nfft, 0),
        ]

    def get_window(self, nfft):
        return window.hann(nfft)

    def get_pretune_block(
        self,
        fft_batch_size,
        nfft,
        samp_rate,
        tune_jitter_hz,
        freq_start,
        freq_end,
        tune_step_hz,
        tune_step_fft,
        skip_tune_step,
        tuning_ranges,
        pretune,
        low_power_hold_down,
        slew_rx_time,
    ):
        # if pretuning, the pretune block will also do the batching.
        if pretune:
            block = self.iqtlabs.retune_pre_fft(
                nfft=nfft,
                samp_rate=int(samp_rate),
                tune_jitter_hz=int(tune_jitter_hz),
                fft_batch_size=fft_batch_size,
                tag="rx_freq",
                freq_start=int(freq_start),
                freq_end=int(freq_end),
                tune_step_hz=tune_step_hz,
                tune_step_fft=tune_step_fft,
                skip_tune_step_fft=skip_tune_step,
                tuning_ranges=tuning_ranges,
                tag_now=self.tag_now,
                low_power_hold_down=low_power_hold_down,
                slew_rx_time=slew_rx_time,
            )
        else:
            # otherwise, the pretuning block will just do batching.
            block = blocks.stream_to_vector(gr.sizeof_gr_complex, fft_batch_size * nfft)
        return block

    def apply_window(self, nfft, fft_batch_size):
        window_constants = [val for val in self.get_window(nfft) for _ in range(2)]
        return blocks.multiply_const_vff(window_constants * fft_batch_size)

    def get_offload_fft_blocks(
        self,
        vkfft,
        fft_batch_size,
        nfft,
        fft_processor_affinity,
    ):
        fft_block = None
        fft_roll = False
        if self.wavelearner:
            fft_block = self.wavelearner.fft(int(fft_batch_size * nfft), nfft, True)
            fft_roll = True
        elif vkfft:
            # VkFFT handles batches by using set_multiple_output(), so we do not need
            # to wrap it.
            fft_block = self.iqtlabs.vkfft(fft_batch_size, nfft, True)
            fft_batch_size = 1
        else:
            logging.warning(
                "using software FFT - may not be deterministic, even with cached wisdon"
            )
            fft_block = fft.fft_vcc(nfft, True, [], True, 1)
            fft_batch_size = 1
        fft_block.set_thread_priority(99)
        fft_block.set_processor_affinity([fft_processor_affinity])

        fft_blocks = [
            self.apply_window(nfft, fft_batch_size),
            fft_block,
        ]
        if fft_batch_size > 1:
            fft_blocks.append(
                blocks.vector_to_stream(gr.sizeof_gr_complex * nfft, fft_batch_size)
            )
        if fft_roll:
            fft_blocks.append(self.iqtlabs.vector_roll(nfft))
        return fft_batch_size, fft_blocks

    def get_dc_blocks(self, dc_block_len, dc_block_long):
        if dc_block_len:
            return [grfilter.dc_blocker_cc(dc_block_len, dc_block_long)]
        return []

    def get_fft_blocks(
        self,
        samp_rate,
        tune_jitter_hz,
        vkfft,
        fft_batch_size,
        nfft,
        freq_start,
        freq_end,
        tune_step_hz,
        tune_step_fft,
        skip_tune_step,
        tuning_ranges,
        pretune,
        fft_processor_affinity,
        low_power_hold_down,
        slew_rx_time,
    ):
        fft_batch_size, fft_blocks = self.get_offload_fft_blocks(
            vkfft,
            fft_batch_size,
            nfft,
            fft_processor_affinity,
        )
        self.retune_pre_fft = self.get_pretune_block(
            fft_batch_size,
            nfft,
            samp_rate,
            tune_jitter_hz,
            freq_start,
            freq_end,
            tune_step_hz,
            tune_step_fft,
            skip_tune_step,
            tuning_ranges,
            pretune,
            low_power_hold_down,
            slew_rx_time,
        )
        return (fft_batch_size, [self.retune_pre_fft] + fft_blocks)

    def start(self):
        super().start()
        self.workaround_start_hook(self)
        logging.info("raw edge and message edge lists follow")
        logging.info(self.edge_list())
        logging.info(self.msg_edge_list())



def get_soapy_source(sdr, sdrargs, samp_rate, center_freq, agc, gain, soapy):
    dev = f"driver={sdr}"
    stream_args = ""
    tune_args = [""]
    settings = [""]
    if sdrargs:
        settings = sdrargs.split(",")
    source = soapy.source(dev, "fc32", 1, "", stream_args, tune_args, settings)
    source.set_sample_rate(0, samp_rate)
    source.set_bandwidth(0, samp_rate)
    if center_freq is not None:
        source.set_frequency(0, center_freq)
        source.set_frequency_correction(0, 0)
        source.set_gain_mode(0, agc)
        source.set_gain(0, gain)
    cmd_port = "cmd"
    return [source], cmd_port



#!/usr/bin/env python3

import re
import os
import gzip
import sigmf
import zstandard
import numpy as np
from gamutrf.utils import SAMPLE_DTYPES, SAMPLE_FILENAME_RE, is_fft

POINTS_RE = re.compile(r"^.+\D([0-9]+)points_.+$")


def get_reader(filename):
    # nosemgrep:github.workflows.config.useless-inner-function
    def gzip_reader(x):
        return gzip.open(x, "rb")

    # nosemgrep:github.workflows.config.useless-inner-function
    def zst_reader(x):
        return zstandard.ZstdDecompressor().stream_reader(
            open(x, "rb"), read_across_frames=True
        )

    def default_reader(x):
        return open(x, "rb")

    if filename.endswith(".gz"):
        return gzip_reader
    if filename.endswith(".zst"):
        return zst_reader

    return default_reader


def parse_filename(filename):
    timestamp = None
    nfft = None
    center_frequency = None
    sample_rate = None
    sample_type = None

    # FFT is always float not matter the original sample type.
    if is_fft(filename):
        sample_type = "raw"
        match = POINTS_RE.match(filename)
        if match is not None:
            nfft = int(match.group(1))

    match = SAMPLE_FILENAME_RE.match(filename)
    if match is not None:
        timestamp, center_frequency, sample_rate, sample_type = (
            int(match.group(1)),
            int(match.group(2)),
            int(match.group(3)),
            match.group(4),
        )

    sample_dtype, sample_type = SAMPLE_DTYPES.get(sample_type, (None, None))
    sample_bits = None
    sample_len = None
    if sample_dtype:
        if is_fft(filename):
            sample_dtype = np.dtype([("i", np.float32), ("q", np.float32)])
        else:
            sample_dtype = np.dtype([("i", sample_dtype), ("q", sample_dtype)])
        sample_bits = sample_dtype[0].itemsize * 8
        sample_len = sample_dtype[0].itemsize * 2
    meta = {
        "filename": filename,
        "center_frequency": center_frequency,
        "sample_rate": sample_rate,
        "sample_dtype": sample_dtype,
        "sample_len": sample_len,
        "sample_type": sample_type,
        "sample_bits": sample_bits,
        "nfft": nfft,
        "timestamp": timestamp,
    }
    return meta


def read_recording(
    filename,
    sample_rate,
    sample_dtype,
    sample_len,
    sample_secs=1.0,
    skip_sample_secs=0,
    max_sample_secs=0,
):
    """Read an I/Q recording and iterate over it, returning 1-D numpy arrays of csingles, of size sample_rate * sample_secs.

    Args:
        filename: str, recording to read.
        sample_rate: int, samples per second
        sample_dtype: numpy.dtype, binary format of original I/Q recording.
        sample_len: int, length of one sample.
        sample_secs: float, number of seconds worth of samples per iteration.
        skip_sample_secs: float, number of seconds worth of samples to skip initially.
        max_sample_secs: float, maximum number of seconds of samples to read (or None for all).
    Returns:
        numpy arrays of csingles.
    """
    read_size = int(sample_rate * sample_secs) * sample_len
    reader = get_reader(filename)
    samples_remaining = 0
    if max_sample_secs:
        samples_remaining = int(sample_rate * max_sample_secs)
    with reader(filename) as infile:
        if skip_sample_secs:
            infile.seek(int(sample_rate * skip_sample_secs) * sample_len)
        while True:
            if max_sample_secs and not samples_remaining:
                break
            sample_buffer = infile.read(read_size)
            buffered_samples = int(len(sample_buffer) / sample_len)
            if buffered_samples == 0:
                break
            if max_sample_secs:
                if buffered_samples <= samples_remaining:
                    samples_remaining -= buffered_samples
                else:
                    buffered_samples = samples_remaining
                    samples_remaining = 0
            x1d = np.frombuffer(
                sample_buffer, dtype=sample_dtype, count=buffered_samples
            )
            yield x1d["i"] + np.csingle(1j) * x1d["q"]


def get_nosigmf_samples(filename):
    meta = parse_filename(filename)
    samples = None
    for samples_buffer in read_recording(
        filename,
        meta["sample_rate"],
        meta["sample_dtype"],
        meta["sample_len"],
        max_sample_secs=None,
    ):
        if samples is None:
            samples = samples_buffer
        else:
            samples = np.concatenate([samples, samples_buffer])
    return filename, samples, meta


def get_samples(filename):
    if not os.path.exists(filename):
        raise FileNotFoundError(filename)
    meta_ext = filename.find(".sigmf-meta")
    if meta_ext == -1:
        return get_nosigmf_samples(filename)

    meta = sigmf.sigmffile.fromfile(filename)
    data_filename = filename[:meta_ext]
    if os.path.splitext(data_filename)[-1] != ".sigmf-data":
        data_filename = data_filename + ".sigmf-data"
    if not os.path.exists(data_filename):
        raise FileNotFoundError(data_filename)

    meta.set_data_file(data_filename)
    samples = meta.read_samples()
    global_meta = meta.get_global_info()
    captures_meta = meta.get_captures()
    center_frequency = None
    timestamp = None
    if captures_meta:
        center_frequency = captures_meta[0].get("core:frequency", None)
        timestamp = captures_meta[0].get("core:datetime", None)
        if timestamp is not None:
            TS_RE = re.compile(r"^([^\.]+)([\.\d]+)Z$")
            ts_match = TS_RE.match(timestamp)
            if ts_match is None:
                print(
                    "invalid SigMF timestamp (does not match %s): %s"
                    % (TS_RE, timestamp)
                )
                timestamp = None
            else:
                secs_str, remainder = ts_match.group(1), float(ts_match.group(2))
                remainder_str = ("%.3f" % remainder)[1:]
                try:
                    timestamp = sigmf.utils.parse_iso8601_datetime(
                        secs_str + remainder_str + "Z"
                    ).timestamp()
                except ValueError as e:
                    print("invalid SigMF timestamp: %s" % e)
    if timestamp is None:
        filename_meta = parse_filename(data_filename)
        timestamp = filename_meta.get("timestamp", timestamp)
    if timestamp is None:
        print("warning: no SigMF or filename timestamp available, using ctime")
        timestamp = os.stat(data_filename).st_ctime
    meta = {
        "sample_rate": global_meta["core:sample_rate"],
        "sample_dtype": global_meta["core:datatype"],
        "sample_len": samples[
            0
        ].itemsize,  # read_samples() always converts to host cf32.
        "center_frequency": center_frequency,
        "timestamp": timestamp,
    }
    return data_filename, samples, meta



#!/usr/bin/python3
import glob
import os
import logging

from gnuradio import iqtlabs
from gamutrf.scan import argument_parser, DYNAMIC_EXCLUDE_OPTIONS
from gamutrf.grscan import grscan
from gamutrf.sample_reader import get_samples


def main():
    logging.basicConfig(level=logging.DEBUG, format="%(asctime)s %(message)s")
    parser = argument_parser()
    parser.add_argument("filename", type=str, help="Recording filename (or glob)")
    options = parser.parse_args()
    outputs = 0
    for filename in glob.glob(options.filename):
        out_dir = os.path.dirname(filename)
        if out_dir == "":
            out_dir = "."
        _data_filename, _samples, meta = get_samples(filename)
        freq_start = int(meta["center_frequency"] - (meta["sample_rate"] / 2))
        scan_args = {
            k: getattr(options, k)
            for k in dir(options)
            if not k.startswith("_")
            and k != "filename"
            and k not in DYNAMIC_EXCLUDE_OPTIONS
        }
        for override_dir in ("inference_output_dir", "sample_dir"):
            override_val = getattr(options, override_dir)
            if not override_val:
                override_val = out_dir
            scan_args[override_dir] = override_val
        scan_args.update(
            {
                "iqtlabs": iqtlabs,
                "freq_end": 0,
                "freq_start": freq_start,
                "samp_rate": int(meta["sample_rate"]),
                "sdr": "file:" + filename,
                "pretune": True,
                "fft_batch_size": 1,
                "low_power_hold_down": False,
                "iq_inference_background": False,
            }
        )
        tb = grscan(**scan_args)
        tb.start()
        tb.wait()
        tb.stop()
        outputs += 1
    logging.info("%u filenames processed from %s", outputs, options.filename)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import json
import logging
import queue
import sys
import threading
import time
import pmt
import zmq

try:
    from gnuradio import gr  # pytype: disable=import-error
    from gamutrf.mqtt_reporter import MQTTReporter
except ModuleNotFoundError as err:  # pragma: no cover
    print(
        "Run from outside a supported environment, please run via Docker (https://github.com/IQTLabs/gamutRF#readme): %s"
        % err
    )
    sys.exit(1)


DELIM = "\n\n"


class inferenceoutput(gr.basic_block):
    def __init__(
        self,
        name,
        zmq_addr,
        mqtt_server,
        compass,
        gps_server,
        use_external_gps,
        use_external_heading,
        external_gps_server,
        external_gps_server_port,
        log_path,
    ):
        self.q = queue.Queue()
        self.running = True
        self.serialno = 0
        self.reporter_thread = threading.Thread(
            target=self.run_reporter_thread,
            args=(
                name,
                zmq_addr,
                mqtt_server,
                gps_server,
                compass,
                use_external_gps,
                use_external_heading,
                external_gps_server,
                external_gps_server_port,
                log_path,
            ),
        )
        self.reporter_thread.start()
        gr.basic_block.__init__(
            self,
            name="inferenceoutput",
            in_sig=None,
            out_sig=None,
        )
        self.message_port_register_in(pmt.intern("inference"))
        self.set_msg_handler(pmt.intern("inference"), self.receive_pdu)

    def receive_pdu(self, pdu):
        self.q.put(json.loads(bytes(pmt.to_python(pmt.cdr(pdu))).decode("utf8")))

    def stop(self):
        self.running = False
        self.reporter_thread.join()

    def run_reporter_thread(
        self,
        name,
        zmq_addr,
        mqtt_server,
        gps_server,
        compass,
        use_external_gps,
        use_external_heading,
        external_gps_server,
        external_gps_server_port,
        log_path,
    ):
        start_time = time.time()
        zmq_context = None
        zmq_pub = None
        if zmq_addr:
            zmq_context = zmq.Context()
            zmq_pub = zmq_context.socket(zmq.PUB)
            zmq_pub.setsockopt(zmq.SNDHWM, 100)
            zmq_pub.setsockopt(zmq.SNDBUF, 65536)
            zmq_pub.bind(zmq_addr)
        mqtt_reporter = None
        if mqtt_server:
            mqtt_reporter = MQTTReporter(
                name=name,
                mqtt_server=mqtt_server,
                gps_server=gps_server,
                compass=compass,
                use_external_gps=use_external_gps,
                use_external_heading=use_external_heading,
                external_gps_server=external_gps_server,
                external_gps_server_port=external_gps_server_port,
            )
        while self.running:
            try:
                item = self.q.get(block=True, timeout=1)
            except queue.Empty:
                continue
            logging.info("inference output %u: %s", self.serialno, item)
            self.serialno += 1
            if zmq_pub is not None:
                zmq_pub.send_string(json.dumps(item) + DELIM, flags=zmq.NOBLOCK)
            if mqtt_reporter is not None:
                mqtt_reporter.publish("gamutrf/inference", item)
                mqtt_reporter.log(log_path, "inference", start_time, item)
            self.q.task_done()



import logging
import os
import sys
import time
from urllib.parse import urlparse
import numpy as np

try:
    import pmt
    from gnuradio import blocks
    from gnuradio import soapy
    from gnuradio import uhd
    from gnuradio import gr
    from gnuradio import iqtlabs
except ModuleNotFoundError as err:  # pragma: no cover
    print(
        "Run from outside a supported environment, please run via Docker (https://github.com/IQTLabs/gamutRF#readme): %s"
        % err
    )
    sys.exit(1)

from gamutrf.ettus_source import get_ettus_source
from gamutrf.soapy_source import get_soapy_source
from gamutrf.sample_reader import get_samples


def null_workaround_start_hook(self):
    return


def airt_workaround_start_hook(self):
    logging.info("applying AIRT workarounds")
    workaround_time = 0.5
    for rate in (15.625e6, self.samp_rate):
        time.sleep(workaround_time)
        self.sources[0].set_sample_rate(0, rate)
        self.sources[0].set_bandwidth(0, rate)


class file_source_tagger(gr.basic_block):
    def __init__(
        self,
        input_file,
        cmd_port,
        nfft,
        tune_step_fft,
    ):
        gr.basic_block.__init__(
            self,
            name="file_source_tagger",
            in_sig=None,
            out_sig=[np.complex64],
        )
        _, self.samples, meta = get_samples(input_file)
        self.timestamp = meta["timestamp"]
        self.center_freq = meta["center_frequency"]
        self.sample_rate = meta["sample_rate"]
        self.n_samples = len(self.samples)
        self.nfft = nfft
        self.tune_step_fft = tune_step_fft
        self.cmds_received = 0
        self.tags_sent = -1
        self.i = 0
        self.message_port_register_in(pmt.intern(cmd_port))
        self.set_msg_handler(pmt.intern(cmd_port), self.handle_cmd)
        self.set_output_multiple(self.nfft)
        self.tagged_interval = self.nfft * self.tune_step_fft
        self.n_samples = (
            int(self.n_samples / self.tagged_interval) * self.tagged_interval
        )
        logging.info("opened %s with %u samples", input_file, self.n_samples)
        self.work_guard = None

    def complete(self):
        return self.i >= self.n_samples

    def handle_cmd(self, _msg):
        self.cmds_received += 1

    def make_rx_time(self, sample_time):
        sample_sec = int(sample_time)
        sample_fsec = sample_time - sample_sec
        pmt_sample_time = pmt.make_tuple(
            pmt.from_long(sample_sec), pmt.from_double(sample_fsec)
        )
        return pmt_sample_time

    def add_tags(self):
        # Ideally, We want to add simulated rx_time tags in a consistent way between
        # runs. However, gnuradio's scheduler does not gaurantee when messages between
        # blocks are delivered. So more or less items may be processed before
        # a simulated tuning request is received by this block, between runs. This
        # will result in different timestamps.
        self.tags_sent += 1
        tag_pos = self.tags_sent * self.tagged_interval
        sample_time = self.timestamp + (tag_pos / float(self.sample_rate))
        logging.info(
            "tag %u at pos %u (nfft item %u), %.1f%%",
            self.tags_sent,
            tag_pos,
            int(tag_pos / self.nfft),
            self.i / self.n_samples * 100,
        )
        self.add_item_tag(
            0,
            tag_pos,
            pmt.intern("rx_freq"),
            pmt.from_double(self.center_freq),
        )
        self.add_item_tag(
            0,
            tag_pos,
            pmt.intern("rx_time"),
            self.make_rx_time(sample_time),
        )

    def general_work(self, input_items, output_items):
        if self.complete():
            # gnuradio will drop all undelivered messages to blocks when our source
            # returns done. cause gnuradio to repeatedly call us when we're done, to
            # give other blocks an opportunity to process undelivered messages.
            if self.work_guard is None:
                self.work_guard = time.time()
                logging.info("file ended, waiting for other blocks to finish")
                return 0
            if time.time() - self.work_guard < 3:
                return 0
            logging.info("complete")
            return -1

        n = min(self.nfft, len(output_items[0]))
        samples = self.samples[self.i : self.i + n]
        c = len(samples)
        self.i += c
        if c < len(output_items[0]):
            zeros = np.zeros(len(output_items[0]) - c, dtype=np.complex64)
            samples = np.append(samples, zeros)
        while (
            not self.complete() and int(self.i / self.tagged_interval) != self.tags_sent
        ):
            self.add_tags()
        output_items[0][:] = samples
        return c


def get_throttle(samp_rate, items):
    return blocks.throttle(gr.sizeof_gr_complex, samp_rate, True, items)


def get_source(
    sdr,
    samp_rate,
    gain,
    nfft,
    tune_step_fft,
    agc=False,
    center_freq=None,
    sdrargs=None,
    soapy_lib=soapy,
    uhd_lib=uhd,
    dc_ettus_auto_offset=True,
):
    logging.info(
        f"initializing SDR {sdr} with sample rate {samp_rate}, gain {gain}, agc {agc}"
    )

    workaround_start_hook = null_workaround_start_hook
    cmd_port = "command"
    sources = []

    url = urlparse(sdr)
    if url.scheme:
        if url.scheme == "file" and os.path.exists(url.path):
            sources = [
                file_source_tagger(url.path, cmd_port, nfft, tune_step_fft),
            ]
        else:
            raise ValueError("unsupported/missing file location")
    elif sdr == "tuneable_test_source":
        freq_divisor = 1e9
        cmd_port = "cmd"
        sources = [
            iqtlabs.tuneable_test_source(0, freq_divisor),
            get_throttle(samp_rate, nfft),
        ]
    elif sdr == "ettus":
        sources = get_ettus_source(
            sdrargs, samp_rate, center_freq, agc, gain, uhd_lib, dc_ettus_auto_offset
        )
    else:
        sources, cmd_port = get_soapy_source(
            sdr, sdrargs, samp_rate, center_freq, agc, gain, soapy_lib
        )
        if sdr == "SoapyAIRT":
            workaround_start_hook = airt_workaround_start_hook
    sources[0].set_thread_priority(99)
    return sources, cmd_port, workaround_start_hook



"""Main entrypoint for GamutRF"""

from gamutrf.compress_dirs import main as compress_dirs_main
from gamutrf.offline import main as offline_main
from gamutrf.scan import main as scan_main


def compress_dirs():
    """Entrypoint for compress_dirs"""
    compress_dirs_main()


def offline():
    """Entrypoint for offline"""
    offline_main()


def scan():
    """Entrypoint for scan"""
    scan_main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import logging
import sys
import pmt
import zmq
import zstandard

try:
    from gnuradio import gr  # pytype: disable=import-error
except ModuleNotFoundError as err:  # pragma: no cover
    print(
        "Run from outside a supported environment, please run via Docker (https://github.com/IQTLabs/gamutRF#readme): %s"
        % err
    )
    sys.exit(1)

DELIM = "\n"


# It would be ideal to just use gnuradio's https://github.com/gnuradio/gnuradio/blob/main/gr-zeromq/lib/pub_msg_sink_impl.cc
# block. Unfortunately, this block calls pmt::serialize(), but we just want to send simple json strings. That means
# a receiver client would need to run pmt::deserialize() which requires an installation of gnuradio.
class pduzmq(gr.basic_block):
    def __init__(
        self,
        zmq_addr,
    ):
        gr.basic_block.__init__(
            self,
            name="pduzmq",
            in_sig=None,
            out_sig=None,
        )
        self.zmq_context = zmq.Context()
        self.zmq_pub = self.zmq_context.socket(zmq.PUB)
        self.zmq_pub.setsockopt(zmq.SNDHWM, 100)
        self.zmq_pub.setsockopt(zmq.SNDBUF, 65536)
        self.zmq_pub.bind(zmq_addr)
        self.message_port_register_in(pmt.intern("json"))
        self.set_msg_handler(pmt.intern("json"), self.receive_pdu)
        self.context = zstandard.ZstdCompressor()

    def stop(self):
        self.zmq_pub.close()

    def receive_pdu(self, pdu):
        item = pmt.to_python(pmt.cdr(pdu)).tobytes().decode("utf8").strip()
        try:
            data = item + DELIM
            data = self.context.compress(data.encode("utf8"))
            self.zmq_pub.send(data, flags=zmq.NOBLOCK)
        except zmq.ZMQError as e:
            logging.error(str(e))



import logging
import signal
import time
import sys
from argparse import ArgumentParser, BooleanOptionalAction

try:
    from gnuradio import iqtlabs  # pytype: disable=import-error
    from gnuradio import eng_notation  # pytype: disable=import-error
    from gnuradio import gr  # pytype: disable=import-error
    from gnuradio.eng_arg import eng_float  # pytype: disable=import-error
    from gnuradio.eng_arg import intx  # pytype: disable=import-error
except ModuleNotFoundError as err:  # pragma: no cover
    print(
        "Run from outside a supported environment, please run via Docker (https://github.com/IQTLabs/gamutRF#readme): %s"
        % err
    )
    sys.exit(1)

from prometheus_client import Gauge
from prometheus_client import start_http_server

from gamutrf.grscan import grscan
from gamutrf.flask_handler import FlaskHandler
from gamutrf.utils import SAMP_RATE, MIN_FREQ, MAX_FREQ

running = True
DYNAMIC_EXCLUDE_OPTIONS = ["apiport", "promport", "updatetimeout"]


def init_prom_vars():
    prom_vars = {
        "freq_start": Gauge("freq_start", "start of scanning range in Hz"),
        "freq_end": Gauge("freq_end", "end of scanning range in Hz"),
        "igain": Gauge("igain", "input gain"),
        "tuneoverlap": Gauge("tuneoverlap", "multiple of samp_rate when retuning"),
        "tune_step_fft": Gauge("tune_step_fft", "tune FFT step (0 is use sweep_sec)"),
        "sweep_sec": Gauge("sweep_sec", "scan sweep rate in seconds"),
        "run_timestamp": Gauge("run_timestamp", "updated when flowgraph is running"),
    }
    return prom_vars


def argument_parser():
    parser = ArgumentParser()
    parser.add_argument(
        "--freq-end",
        dest="freq_end",
        type=eng_float,
        default=eng_notation.num_to_str(MAX_FREQ),
        help="Set freq_end [default=%(default)r] (if 0, configure stare mode at --freq-start)",
    )
    parser.add_argument(
        "--freq-start",
        dest="freq_start",
        type=eng_float,
        default=eng_notation.num_to_str(MIN_FREQ),
        help="Set freq_start [default=%(default)r]",
    )
    parser.add_argument(
        "--igain",
        dest="igain",
        type=intx,
        default=0,
        help="Set igain[default=%(default)r]",
    )
    parser.add_argument(
        "--samp-rate",
        dest="samp_rate",
        type=eng_float,
        default=eng_notation.num_to_str(SAMP_RATE),
        help="Set samp_rate [default=%(default)r]",
    )
    parser.add_argument(
        "--tune-jitter-hz",
        dest="tune_jitter_hz",
        type=eng_float,
        default=0,
        help="If > 0, apply a random tuning jitter in Hz from center frequency",
    )
    parser.add_argument(
        "--sweep-sec",
        dest="sweep_sec",
        type=float,
        default=30,
        help="Set sweep_sec [default=%(default)r] - ignored if --tune-dwell-ms > 0 or --tune-step-fft > 0",
    )
    parser.add_argument(
        "--tune-dwell-ms",
        dest="tune_dwell_ms",
        type=float,
        default=0,
        help="Set tune dwell time in ms [default=%(default)r] - ignored if --tune-step-fft > 0",
    )
    parser.add_argument(
        "--tune-step-fft",
        dest="tune_step_fft",
        type=int,
        default=0,
        help="tune FFT step [default=%(default)r] - if 0, use --tune-dwell-ms (if > 0) or --sweep-sec (if > 0)",
    )
    parser.add_argument(
        "--skip-tune-step",
        dest="skip_tune_step",
        type=int,
        default=0,
        help="skip FFT samples on retune [default=%(default)r]",
    )
    parser.add_argument(
        "--sample_dir",
        dest="sample_dir",
        type=str,
        default="",
        help="where to write samples/FFT points",
    )
    parser.add_argument(
        "--write_samples",
        dest="write_samples",
        default=0,
        type=int,
        help="if > 0, write FFT points/raw samples to --sample_dir",
    )
    parser.add_argument(
        "--write_fft_points",
        dest="write_fft_points",
        default=False,
        action=BooleanOptionalAction,
        help="Write FFT points to --sample_dir if write_samples > 0",
    )
    parser.add_argument(
        "--nfft",
        dest="nfft",
        type=int,
        default=1024,
        help="FFTI size [default=%(default)r]",
    )
    parser.add_argument(
        "--fft_zmq_addr",
        dest="fft_zmq_addr",
        type=str,
        default="0.0.0.0",  # nosec
        help="Serve FFT results from this address over ZMQ",
    )
    parser.add_argument(
        "--fft_zmq_port",
        dest="fft_zmq_port",
        type=int,
        default=10000,
        help="if > 0, serve FFT results to this port over ZMQ",
    )
    parser.add_argument(
        "--inference_batch",
        dest="inference_batch",
        type=int,
        default=5,
        help="How many inference requests to make at once",
    )
    parser.add_argument(
        "--inference_addr",
        dest="inference_addr",
        type=str,
        default="0.0.0.0",  # nosec
        help="Log inference results to this address",
    )
    parser.add_argument(
        "--inference_port",
        dest="inference_port",
        type=int,
        default=10001,
        help="Log inference results to this port",
    )
    parser.add_argument(
        "--iq_zmq_addr",
        dest="iq_zmq_addr",
        type=str,
        default="0.0.0.0",  # nosec
        help="Serve scanner I/Q stream and tags from this address over ZMQ",
    )
    parser.add_argument(
        "--iq_zmq_port",
        dest="iq_zmq_port",
        type=int,
        default=10002,
        help="If > 0, Serve scanner I/Q stream and tags from this port over ZMQ",
    )
    parser.add_argument(
        "--inference_text_color",
        dest="inference_text_color",
        type=str,
        default="white",
        help="CSS3 color name for inference text (https://www.w3.org/wiki/CSS/Properties/color/keywords)",
    )
    parser.add_argument(
        "--promport",
        dest="promport",
        type=int,
        default=9000,
        help="Prometheus client port",
    )
    parser.add_argument(
        "--apiport",
        dest="apiport",
        type=int,
        default=9001,
        help="API server port",
    )
    parser.add_argument(
        "--rotate_secs",
        dest="rotate_secs",
        type=int,
        default=300,
        help="rotate storage directories every N seconds",
    )
    parser.add_argument(
        "--sdr",
        dest="sdr",
        type=str,
        default="ettus",
        help="SDR to use (ettus, bladerf, or lime)",
    )
    parser.add_argument(
        "--sdrargs",
        dest="sdrargs",
        type=str,
        default="",
        help="extra args to pass to SDR driver",
    )
    parser.add_argument(
        "--updatetimeout",
        dest="updatetimeout",
        type=int,
        default=10,
        help="seconds to wait for healthy freq updates",
    )
    parser.add_argument(
        "--tuneoverlap",
        dest="tuneoverlap",
        type=float,
        default=0.85,
        help="multiple of samp_rate when retuning",
    )
    parser.add_argument(
        "--bucket_range",
        dest="bucket_range",
        type=float,
        default=0.85,
        help="what proportion of FFT buckets to use",
    )
    parser.add_argument(
        "--db_clamp_floor",
        dest="db_clamp_floor",
        type=float,
        default=-200,
        help="clamp dB output floor",
    )
    parser.add_argument(
        "--db_clamp_ceil",
        dest="db_clamp_ceil",
        type=float,
        default=50,
        help="clamp dB output ceil",
    )
    parser.add_argument(
        "--dc_block_len",
        dest="dc_block_len",
        type=int,
        default=0,
        help="if > 0, use dc_block_cc filter with length",
    )
    parser.add_argument(
        "--dc_block_long",
        dest="dc_block_long",
        action="store_true",
        help="Use dc_block_cc long form",
    )
    parser.add_argument(
        "--ettus-dc-offset",
        dest="dc_ettus_auto_offset",
        default=True,
        action=BooleanOptionalAction,
        help="Use Ettus DC Auto offset",
    )
    parser.add_argument(
        "--inference_min_confidence",
        dest="inference_min_confidence",
        type=float,
        default=0.25,
        help="minimum confidence score to plot",
    )
    parser.add_argument(
        "--inference_min_db",
        dest="inference_min_db",
        type=float,
        default=-200,
        help="run inference over minimum dB power",
    )
    parser.add_argument(
        "--inference_model_server",
        dest="inference_model_server",
        type=str,
        default="",
        help="torchserve model server inference API address (e.g. localhost:1234)",
    )
    parser.add_argument(
        "--iq_inference_background",
        dest="iq_inference_background",
        type=bool,
        default=True,
        help="if True do I/Q inference in the background",
    )
    parser.add_argument(
        "--iq_inference_squelch_db",
        dest="iq_inference_squelch_db",
        type=float,
        default=None,
        help="if not None, preface I/Q inference with power squelch dB threshold",
    )
    parser.add_argument(
        "--iq_inference_squelch_alpha",
        dest="iq_inference_squelch_alpha",
        type=float,
        default=1e-4,
        help="I/Q inference power squelch alpha",
    )
    parser.add_argument(
        "--iq_inference_model_name",
        dest="iq_inference_model_name",
        type=str,
        default="",
        help="torchserve model name",
    )
    parser.add_argument(
        "--iq_inference_model_server",
        dest="iq_inference_model_server",
        type=str,
        default="",
        help="torchserve model server inference API address (e.g. localhost:1234)",
    )
    parser.add_argument(
        "--inference_model_name",
        dest="inference_model_name",
        type=str,
        default="",
        help="torchserve model name (e.g. mini2_snr)",
    )
    parser.add_argument(
        "--inference_output_dir",
        dest="inference_output_dir",
        type=str,
        default="",
        help="directory for inference output",
    )
    parser.add_argument(
        "--tuning_ranges",
        dest="tuning_ranges",
        type=str,
        default="",
        help="tuning ranges (overriding freq_start and freq_end)",
    )
    parser.add_argument(
        "--description",
        dest="description",
        type=str,
        default="",
        help="optional text description to provide with scanner updates",
    )
    parser.add_argument(
        "--scaling",
        dest="scaling",
        type=str,
        default="spectrum",
        help="""Same as --scaling parameter in scipy.signal.spectrogram().
        Selects between computing the power spectral density ('density') 
        where `Sxx` has units of V**2/Hz and computing the power
        spectrum ('spectrum') where `Sxx` has units of V**2, if `x`
        is measured in V and `fs` is measured in Hz. Defaults to
        'spectrum'.""",
    )
    parser.add_argument(
        "--sigmf",
        dest="sigmf",
        default=True,
        action=BooleanOptionalAction,
        help="add sigmf meta file",
    )
    parser.add_argument(
        "--fft_batch_size",
        dest="fft_batch_size",
        type=int,
        default=256,
        help="offload FFT batch size",
    )
    parser.add_argument(
        "--fft_processor_affinity",
        dest="fft_processor_affinity",
        type=int,
        default=0,
        help="FFT processor affinity",
    )
    parser.add_argument(
        "--vkfft",
        dest="vkfft",
        default=True,
        action=BooleanOptionalAction,
        help="use VkFFT (ignored if wavelearner available)",
    )
    parser.add_argument(
        "--pretune",
        dest="pretune",
        default=False,
        action=BooleanOptionalAction,
        help="use pretuning",
    )
    parser.add_argument(
        "--tag-now",
        dest="tag_now",
        default=True,
        action=BooleanOptionalAction,
        help="send tag:now command when retuning",
    )
    parser.add_argument(
        "--compass",
        dest="compass",
        default=False,
        action=BooleanOptionalAction,
        help="use compass",
    )
    parser.add_argument(
        "--use_external_gps",
        dest="use_external_gps",
        default=False,
        action=BooleanOptionalAction,
        help="Use external Pixhawk/MAVLINK GPS",
    )
    parser.add_argument(
        "--use_external_heading",
        dest="use_external_heading",
        default=False,
        action=BooleanOptionalAction,
        help="Use external (Pixhawk/MAVLINK) heading",
    )
    parser.add_argument(
        "--external_gps_server",
        dest="external_gps_server",
        default="",
        type=str,
        help="server to query for external GPS data",
    )
    parser.add_argument(
        "--external_gps_server_port",
        dest="external_gps_server_port",
        default="8888",
        type=str,
        help="server port to query for external GPS data",
    )
    parser.add_argument(
        "--mqtt_server",
        help="MQTT server to report RSSI",
        default="",
        type=str,
    )
    parser.add_argument(
        "--gps_server",
        help="GPS Server to get lat,long, and heading",
        default="",
        type=str,
    )
    parser.add_argument(
        "--low_power_hold_down",
        help="Gate samples on low power sample period (recommended for Ettus)",
        default=True,
        action=BooleanOptionalAction,
    )
    parser.add_argument(
        "--slew_rx_time",
        help="slew rx_time based on hold down time (recommended for Ettus)",
        default=True,
        action=BooleanOptionalAction,
    )
    parser.add_argument(
        "--iq_power_inference",
        help="send power as well as samples for I/Q inference)",
        default=False,
        action=BooleanOptionalAction,
    )
    parser.add_argument(
        "--n_image",
        dest="n_image",
        type=int,
        default=0,
        help="if > 1, only log 1/n images",
    )
    parser.add_argument(
        "--n_inference",
        dest="n_inference",
        type=int,
        default=0,
        help="if > 1, only send 1/n images for inference",
    )
    parser.add_argument(
        "--colormap",
        dest="colormap",
        type=int,
        default=20,
        help="""OpenCV colormap for spectrograms. See
        https://docs.opencv.org/4.x/d3/d50/group__imgproc__colormap.html""",
    )
    parser.add_argument(
        "--peak_fft_range",
        dest="peak_fft_range",
        type=int,
        default=0,
        help="If > 0, find peak of FFT averages over this many rows",
    )
    return parser


def check_options(options):
    if options.samp_rate % options.nfft:
        print("NFFT should be a factor of sample rate")

    if options.freq_end:
        if options.freq_start > options.freq_end:
            return "freq_start must be less than freq_end"

        if options.freq_end > 6e9:
            return "freq_end must be less than 6GHz"

    if options.freq_start < 10e6:
        return "freq_start must be at least 10MHz"

    if options.write_samples and not options.sample_dir:
        return "Must provide --sample_dir when writing samples/points"

    if options.scaling not in ["spectrum", "density"]:
        return "scaling must be 'spectrum' or 'density'"

    iq_inference = options.iq_inference_model_server and options.iq_inference_model_name
    if iq_inference and not options.pretune:
        return "I/Q inference requires pretune"

    return ""


def sig_handler(_sig=None, _frame=None):
    global running
    running = False


def run_loop(options, prom_vars, wavelearner):
    reconfigures = 0
    global running
    running = True

    signal.signal(signal.SIGINT, sig_handler)
    signal.signal(signal.SIGTERM, sig_handler)

    handler = FlaskHandler(options, check_options, DYNAMIC_EXCLUDE_OPTIONS)
    handler.start()

    while running:
        for var in prom_vars.keys():
            if hasattr(handler.options, var):
                prom_vars[var].set(getattr(handler.options, var))
        scan_args = {
            "iqtlabs": iqtlabs,
            "wavelearner": wavelearner,
        }
        scan_args.update(
            {
                k: getattr(handler.options, k)
                for k in dir(handler.options)
                if not k.startswith("_") and not k in DYNAMIC_EXCLUDE_OPTIONS
            }
        )
        tb = grscan(**scan_args)
        tb.start()
        while running and reconfigures == handler.reconfigures:
            idle_time = 1
            prom_vars["run_timestamp"].set(time.time()),
            time.sleep(idle_time)

        while reconfigures != handler.reconfigures:
            reconfigures = handler.reconfigures

        tb.stop()
        tb.wait()
        del tb


def main():
    logging.basicConfig(level=logging.DEBUG, format="%(asctime)s %(message)s")
    options = argument_parser().parse_args()
    if gr.enable_realtime_scheduling() != gr.RT_OK:
        print("Warning: failed to enable real-time scheduling.")
    results = check_options(options)
    if results:
        print(results)
        sys.exit(1)

    wavelearner = None
    try:
        import wavelearner as wavelearner_lib  # pytype: disable=import-error

        wavelearner = wavelearner_lib
        print("using wavelearner")
    except ModuleNotFoundError:
        print("wavelearner not available")

    prom_vars = init_prom_vars()
    start_http_server(options.promport)

    run_loop(options, prom_vars, wavelearner)
